/*=============================================================================+
|                                                                              |
| Copyright 2008                                                               |
| Acrospeed Inc. All right reserved.                                           |
|                                                                              |
+=============================================================================*/
/*! 
*   \file cml_eth.c
*   \brief  Camelot ethernet driver.
*   \author Acrospeed
*/

/*=============================================================================+
| Included Files
+=============================================================================*/
#include  <sys/types.h>
#include  <cyg/infra/cyg_type.h>
#include  <cyg/hal/hal_arch.h>
#include  <cyg/hal/hal_io.h>
#include  <cyg/hal/hal_cache.h>
#include  <cyg/infra/diag.h>
#include  <cyg/hal/drv_api.h>
#include  <cyg/io/eth/netdev.h>
#include  <cyg/io/eth/eth_drv.h>
#include  <cml_eth.h>
#include  <cml_hnat.h>
//#include  <cdb_api.h>

#include <pkgconf/system.h>
#ifdef CYGPKG_NET_LWIP
#include  <network.h>
#else
#include  <net/if.h>
#include  <sys/ioctl.h>
#include  <sys/socket.h>
#include  <sys/mbuf.h>
#endif

/*=============================================================================+
| Define
+=============================================================================*/
/// @cond DOXYGEN
//#define ETH_DEBUG
/// @endcond

/*=============================================================================+
| Variables
+=============================================================================*/

cyg_handle_t    cml_eth_int_handle;	
cyg_interrupt   cml_eth_int_obj;

static cme_vif cml_eth0_priv_data = { unit: 0, };

ETH_DRV_SC(cml_eth_sc0, &cml_eth0_priv_data, "eth0",
           cml_eth_start, cml_eth_stop, cml_eth_control, cml_eth_cansend, cml_eth_send,
           cml_eth_recv, cml_eth_deliver, cml_eth_poll, cml_eth_int_vector);

NETDEVTAB_ENTRY(cml_eth_netdev0, "cml_eth0", cml_eth_drv_init, &cml_eth_sc0); 

static cme_vif cml_eth1_priv_data = { unit: 1, };

ETH_DRV_SC(cml_eth_sc1, &cml_eth1_priv_data, "eth1",
           cml_eth_start, cml_eth_stop, cml_eth_control, cml_eth_cansend, cml_eth_send,
           cml_eth_recv, cml_eth_deliver, cml_eth_poll, cml_eth_int_vector);

NETDEVTAB_ENTRY(cml_eth_netdev1, "cml_eth1", cml_eth_drv_init, &cml_eth_sc1);

/// @endcond
/*=============================================================================+
| Function Prototypes
+=============================================================================*/
/// @cond DOXYGEN

static void cml_eth_start(struct eth_drv_sc *sc, unsigned char *eaddr, int flags);
static void cml_eth_stop(struct eth_drv_sc *sc);
static int cml_eth_control(struct eth_drv_sc *sc, unsigned long key, void *args, int arglen);
static int cml_eth_cansend(struct eth_drv_sc *sc);
static void cml_eth_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list,	int sg_len);
static void cml_eth_deliver(struct eth_drv_sc *sc);
static void cml_eth_poll(struct eth_drv_sc *sc);
static int cml_eth_int_vector(struct eth_drv_sc *sc);
static bool cml_eth_drv_init(struct cyg_netdevtab_entry *pnde);

static void cml_eth_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len, int total_len, unsigned long key) __attribute__ ((section(".iram")));
static unsigned int cml_eth_isr(cyg_vector_t vec, cyg_addrword_t data) __attribute__ ((section(".iram")));
static void cml_eth_dsr(cyg_vector_t vec, cyg_ucount32 count, cyg_addrword_t data) __attribute__ ((section(".iram")));

int cml_eth_os_request_irq(int irq);

void cml_eth_flowctrl(int onoff);
void cml_debug_printf(int type, char *fmt, ...);

/// @endcond
/*=============================================================================+
| Extern Function/Variables	
+=============================================================================*/

/*=============================================================================+
| Functions
+=============================================================================*/
void cml_debug_printf(int type, char *fmt, ...)
{
#ifdef ETH_DEBUG
	
	va_list ap;
	
	if (!(type & DEBUG_MASK))
		return ;			
	va_start(ap, fmt);
	diag_printf(fmt, ap);
	va_end(ap);	
#endif	
}	

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *     cml_eth_isr:
 *
 *      \brief device interrupt handler
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/
static unsigned int cml_eth_isr(cyg_vector_t vec, cyg_addrword_t data)
{
	cml_eth_int_disable();
	DBG_MSG(DEBUG_ISR, "cml_eth_isr\n");

	cyg_drv_interrupt_acknowledge(vec);
	return CYG_ISR_CALL_DSR;
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *      cml_eth_dsr:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_dsr(cyg_vector_t vec, cyg_ucount32 count, cyg_addrword_t data)
{
	DBG_MSG(DEBUG_DSR, "cml_eth_dsr main=%x\n", pcme->sc_main);

	if (pcme->sc_main)
		eth_drv_dsr(vec, count, (cyg_addrword_t) pcme->sc_main);
}

/// @endcond

/*!-----------------------------------------------------------------------------
 *  cml_eth_drv_init:
 * \brief This function attaches the Ethernet interface to the OS. It also calls various driver 
 *        initialization functions to allocate drivers resources and initialize the device.
 *
 * \param pnde: the pointer of the network device.
 *
 * \return TRUE: Success
 * \return FALSE: Failure
 +----------------------------------------------------------------------------*/

int cml_eth_os_request_irq(int irq)
{
	cyg_drv_interrupt_create( irq, 0, (CYG_ADDRWORD)pcme,
		cml_eth_isr, cml_eth_dsr, &cml_eth_int_handle, &cml_eth_int_obj );

	cyg_drv_interrupt_attach(cml_eth_int_handle);
	cyg_drv_interrupt_configure(irq, 1, 0 );
}

void cml_eth_free_txbd(void *buf)
{
eth_drv_funs.tx_done(&cml_eth_sc0, buf, 0);
}
static bool cml_eth_drv_init(struct cyg_netdevtab_entry *pnde)
{
	struct eth_drv_sc *sc;
	cme_vif *vif;
	int unit;

	sc = (struct eth_drv_sc *)(pnde->device_instance);
	vif = (cme_vif *) sc->driver_private;
	unit = vif->unit;
	
	DBG_MSG(DEBUG_INFO, "cml_eth_drv_init: pnde=%x, sc=%x, unit=%d\n",
			(unsigned int) pnde, (unsigned int) sc, unit);
	
	if ((pcme == 0)  && cml_eth_hw_init() != 0)
	{
		DBG_MSG(DEBUG_ERR, "cml_eth_drv_init: failed hw!\n");
		return false;
	}
	vif->sc = sc;

	if (cml_eth_add_vif(unit, vif) != 0)
	{
		DBG_MSG(DEBUG_ERR, "cml_eth_drv_init: failed vif!\n");
		return false;
	}

	HAL_READ_ETHMAC(unit, vif->macaddr);
	cml_eth_setmac(unit, vif->macaddr);
	
	(sc->funs->eth_drv->init)(sc, vif->macaddr);

	return true;
	
}
/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *       cml_eth_start:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_start(struct eth_drv_sc *sc, unsigned char *eaddr, int flags)
{
	cme_vif *vif;
	
	vif = (cme_vif *)sc->driver_private;
	
	CYG_ASSERTC(vif->sc == sc);
	
	DBG_MSG(DEBUG_INFO, "cml_eth%d: start\n", vif->unit);

	if (!(vif->flags & VIF_ACTIVE))
	{
		int disable = !pcme->vifs_map;
		vif->flags |= VIF_ACTIVE;
		cml_eth_vif_start(vif);
		if (disable)
			cyg_drv_interrupt_unmask(pcme->irq);
	}
	
	cml_eth_rx_start();
	cml_eth_tx_start();
}

/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *       cml_eth_stop:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_stop(struct eth_drv_sc *sc)
{
	cme_vif *vif;
	
	vif = (cme_vif *)sc->driver_private;
	
	CYG_ASSERTC(vif->sc == sc);

	DBG_MSG(DEBUG_INFO, "cml_eth_stop(%d)\n", vif->unit);

	if (vif->flags & VIF_ACTIVE)
	{
		vif->flags &= ~VIF_ACTIVE;
		cml_eth_vif_stop(vif);
		if (!pcme->vifs_map)
			cyg_drv_interrupt_mask(pcme->irq);
	}
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *        cml_eth_deliver:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_deliver(struct eth_drv_sc *sc)
{
	unsigned int int_status;
	
	int_status = MREG(MIS);

	MREG(MIC) = int_status;
	DBG_MSG(DEBUG_DELIVER, "cml_eth_deliver. istatus=%x\n", int_status);
	do {
#if 0
		if (int_status & MIS_SW)
		{
			cml_eth_swint();
		}
#endif		
		if (int_status & MIS_SQE)
		{
			//DBG_MSG(DEBUG_INIT, "SQE\n\n");
			MREG(MAUTO)|=(1<<2);
		}
		if (int_status & MIS_HQE)
		{
			//DBG_MSG(DEBUG_INIT, "HQE\n\n");
		}

		if (int_status & MIE_BE)
		{
			//DBG_MSG(DEBUG_INIT, "BE\n\n");
			MREG(MAUTO)|=(1<<0);
		}

		if (int_status & MIS_TX)
		{
			cml_eth_tx_event(&pcme->tx, 0);
		}

		if (int_status & MIS_RX)
		{
			cml_eth_rx_event(&pcme->rx);
		}
		int_status=MREG(MIS);
		MREG(MIC)=int_status;

//	} while(0);
	} while (int_status & MIM_INIT);
	
	cml_eth_int_enable();
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *        cml_eth_cansend:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static int cml_eth_cansend(struct eth_drv_sc *sc)
{
	int free = 0;
	
	cme_vif *vif;
	
	vif = (cme_vif *) sc->driver_private;
	DBG_MSG(DEBUG_TXMORE, "cml_eth_cansend(%d)\n", vif->unit);

	if (vif->flags & VIF_ACTIVE)
		free = pcme->tx.tx_free;
		
	return free;
}
/// @endcond

/*!-----------------------------------------------------------------------------
 *  cml_eth_send:
 * \brief This function transmits the given frame into the DMA engine of the 
 *				Ethernet controller. If the transmit data path is enable, the frame 
 *				will be transmit at once.
 *
 * \param sc: the pointer for the soft control of Ethernet driver.
 * \param sg_list: the packet which is specified via a scatter-gather list.
 * \param sg_len: number of packet scatter in the list.
 * \param total_len: the packet total length.
 * \param key: the packet address for upper layer release.
 *
 * \return None
 +----------------------------------------------------------------------------*/

static void cml_eth_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list,
		int sg_len, int total_len, unsigned long key)
{
	cme_vif *vif = (cme_vif *) sc->driver_private;
	TDMA *tdma = &pcme->tx;
	volatile DES *ptxd, *ptxd_0;
	void *u;
	int i, desc_cnt;
	unsigned long w0,w1;

	CYG_ASSERT(vif->sc == sc, "Invalid device");

	DBG_MSG(DEBUG_TX, "cml_eth_send(%d) key=%08lx, len=%d\n",
			 vif->unit, key, total_len);
	if(tdma->tx_free == 0)
	{ 
		cml_eth_tx_event(tdma, 1);
		if(tdma->tx_free == 0)
			diag_printf("%s, tdma->tx_free:%d\n", __func__, tdma->tx_free);
	}
	if (tdma->tx_free == 0 || total_len > 1518)
		goto drop_tx;

	tdma->tx_pkt ++;

	if(1 == sg_len) 
		cml_hnat_learn_mac((char *)sg_list[0].buf, 1, RX_VIDX(vif->vlan_ctrl));

	if ((sg_len > TX_MAX_BUFCNT) || (sg_len > tdma->tx_free))
	{
		DBG_MSG(DEBUG_INFO, "tx frag key=%08lx, sg_list=%x, sg_len=%d\n", key, (unsigned int)sg_list, sg_len);
		if (!(u=(void*)eth_drv_alloc_buf()))
		{
			diag_printf("%s: no buf for joint\n",__func__);
			goto drop_tx;
		}
	
		eth_drv_sg_copy(u, NBUF_LEN, sg_list, sg_len);
		sg_list[0].buf = u;
		sg_list[0].len = total_len;
		sg_len = 1;		
		sc->funs->eth_drv->tx_done( sc, key, 0 );
		key = 0;
	}
	if (1)
	{

		for (i=0; i< sg_len; i++)
		{
			FLUSH_TX(sg_list[i].buf, sg_list[i].len);
#if 0
			if (sg_len>1)
			{
				diag_printf("tx sg%d,l=%d\n", i, sg_list[i].len);
				//diag_dump_buf(sg_list[i].buf, sg_list[i].len);
			}
#endif
		}

		ptxd_0 = &tdma->ptxd[tdma->tx_head];
		ptxd_0->w1 = virtophy(sg_list[0].buf) | (vif->vlan_ctrl);
		ptxd_0->w0 = (DS0_END_OF_RING &ptxd_0->w0 ) | (sg_list[0].len<<DS0_LEN_S);
		if ( 1== sg_len)
		{
			ptxd_0->w0 |= DS0_TX_OWNBIT |DS0_DF;
		}
		else
		{
			for (i=1; i<sg_len; i++)
			{
				--tdma->tx_free;
				if ( ++tdma->tx_head == tdma->tx_max)
					tdma->tx_head = 0;
				ptxd = &tdma->ptxd[tdma->tx_head];
				ptxd->w1 = virtophy(sg_list[i].buf);
				ptxd->w0 = (ptxd->w0 & DS0_END_OF_RING)
					| (sg_list[i].len << DS0_LEN_S) | DS0_TX_OWNBIT |DS0_M |DS0_DF;
			}
			ptxd->w0 &= ~DS0_M;
			ptxd_0->w0 |= DS0_TX_OWNBIT | DS0_M |DS0_DF ;
		}
		if (key)
		{
			tdma->tx_bd[tdma->tx_head].txdone = cml_eth_free_txbd;
			tdma->tx_bd[tdma->tx_head].buf = (void *) key;		
		}
		else
		{
			// no mbuf header, only cluster buffer
			tdma->tx_bd[tdma->tx_head].buf = u;
		}
		
		--tdma->tx_free;
		if (++tdma->tx_head == tdma->tx_max)
			tdma->tx_head = 0;
	}
	
	MREG(MTT) = 1;
	return;

drop_tx:
	DBG_MSG(DEBUG_TX, "cml_eth_send(%d) drop frame %08lx\n",
			 vif->unit, key);
	tdma->tx_drop++;
	(sc->funs->eth_drv->tx_done)( sc, key, 0 );
	return;
}


/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *        cml_eth_recv:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list,
		int sg_len)
{
	// SHOULD NOT BE HERE
}

/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *        cml_eth_control:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static int cml_eth_control(struct eth_drv_sc *sc, unsigned long cmd, void *args, int arglen)
{
	cme_vif *vif = (cme_vif *) sc->driver_private;
	
	DBG_MSG(DEBUG_IOCTL, "cml_eth%d: cmd=%08lx\n", vif->unit, cmd);
	switch(cmd)
	{
	case ETH_DRV_SET_MAC_ADDRESS:
		if ( 6 != arglen )
			return -1;
		cml_eth_setmac(vif->unit, (unsigned char *)args);
		return 0;

#ifdef ETH_DRV_SET_MC_ALL
    case ETH_DRV_SET_MC_ALL:
         return 0;
#endif

#ifdef ETH_DRV_SET_MC_LIST
	case ETH_DRV_SET_MC_LIST:
		return 0;
#endif

#ifdef SIOCGPHY
    	case SIOCGPHY:
		{
			struct ifreq *ifr= (struct ifreq *)args;
			unsigned int *ret = (unsigned int *)ifr->ifr_ifru.ifru_data;

			short status = cml_eth_mdio_rd(PHY_WAN_PORT, 1);
			
			DBG_MSG(DEBUG_PHY, "wan port status=%x\n", (unsigned int) status);
			if (ret)
				*ret =  ( status & (1<<2) )?  1 : 0 ;
			return 0;
		}
		break;
#endif

#ifdef  SIOCSFC
	case SIOCSFC:
		{
			struct ifreq *ifr= (struct ifreq *)args;

			if((int)ifr->ifr_ifru.ifru_data)
			{
				diag_printf("~~~~~~~~SIOCSFC on\n");
				pcme->flowctrl_mode = 1;
				pcme->flowctrl_state = 0;
				MREG(MFC1) = 1;
			}
			else
			{
				diag_printf("~~~~~~~~SIOCSFC off\n");
				pcme->flowctrl_mode = 0;
				pcme->flowctrl_state = 0;
				/* toggle the bit */
				MREG(MFC1) = 1;
			}
		}
		break;
#endif
#ifdef SIOCSBRG
	case SIOCSBRG:
		{
			struct ifreq *ifr= (struct ifreq *)args;
			int val=*((int *)ifr->ifr_ifru.ifru_data);
			
			if(	val == 1 ) //enable bridge
			{
				diag_printf("!!!!enable bridge!!!!\n");
				cml_eth_bridge_enable();
			}
			else	//disable bidge
			{
				//To be done
				cml_eth_switch_init();
			}		
			
		}
		break;
#endif

	default:
		return -1;
	}
	return 0;
}

/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *        cml_eth_poll:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static void cml_eth_poll(struct eth_drv_sc *sc)
{
	DBG_MSG(DEBUG_DELIVER, "cml_eth_poll.\n");
	cml_eth_deliver(sc);
}

/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *      cml_eth_int_vector:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

static int cml_eth_int_vector(struct eth_drv_sc *sc)
{
	return (int)pcme->irq;
}

/// @endcond


