/*=============================================================================+
|                                                                              |
| Copyright 2008                                                               |
| Acrospeed Inc. All right reserved.                                           |
|                                                                              |
+=============================================================================*/
/*! 
*   \file cml_eth.c
*   \brief  Camelot ethernet driver.
*   \author Acrospeed
*/

/*=============================================================================+
| Included Files
+=============================================================================*/
#include <cml_eth.h>
#include <cml_hnat.h>

struct eth_drv_sg 
{
    void * buf;
    int len;
};

void cml_eth_rx_event(RDMA *rdma) __attribute__ ((section(".iram")));
void cml_eth_tx_event(TDMA *tdma, int wait_free) __attribute__ ((section(".iram")));
/*=============================================================================+
| Variables	
+=============================================================================*/

static BDS g_txbd[NUM_TX_DESC] __attribute__ ((aligned (0x8)));
// descriptor must be aligned to 0x20 to optimize performance
static unsigned char g_desc[DESC_SZ] __attribute__ ((aligned (0x20)));

cml_eth cml_eth_this;
cml_eth *pcme = 0;
/*=============================================================================+
| Functions
+=============================================================================*/



/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *  cml_eth_mdio:
 *      \brief read/write phy register.
 *	\param phy: 
 *	\param reg: 
 +----------------------------------------------------------------------------*/
void cml_eth_mdio(short phy,short reg,unsigned short *pval,int wr)
{
    unsigned long v;
    
    v=(2<<30)|(phy<<24)|(reg<<16);
    if (wr)
	v |= (1<<30)|*pval;
    MREG(MPHY) = v;
    hal_delay_us(5); /* delay for h/w, DON'T remove*/
    for (;;)
    {
	if ((1<<31)&(v=MREG(MPHY)) )
	{
		break;
	}
    }
    if (!wr)
    	*pval=(v&0xffff);
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 * function: cml_eth_mdio_rd
 *
 *      \brief read phy register.
 *	\param phy: 
 *	\param reg: 
 *      \return 0 is ready, -1 is failed.
 +----------------------------------------------------------------------------*/
unsigned short cml_eth_mdio_rd(short phy,short reg)
{
	unsigned short rd;
	cml_eth_mdio(phy,reg,&rd,0);
	return rd;
}
/// @endcond
/*!-----------------------------------------------------------------------------
 *  cml_eth_mdio_wr:
 *      \brief write phy register.
 *	\param phy: 
 *	\param reg: 
 +----------------------------------------------------------------------------*/
void cml_eth_mdio_wr(short phy,short reg,unsigned short val)
{
	cml_eth_mdio(phy,reg,&val,1);
	if (DEBUG_MASK & DEBUG_PHY)
	{
		unsigned short rd;
		rd=cml_eth_mdio_rd(phy,reg);

        	DBG_MSG(DEBUG_PHY, "phy%02x,reg%02x,wr:%04x\n", phy,(int)reg,(unsigned int)val);
        	if (rd!=val)
	    		diag_printf("cml_eth_mdio_wr(%02x,%02x,%04x)==> not match, rd:%x\n\n", phy, (int)reg, (unsigned int) val, (unsigned int)rd);
	}
}
/// @endcond

/*!-----------------------------------------------------------------------------
 *  cml_eth_phy_up:
 *      \brief power up of lan/wan.
 *	\param phy: 
 *	\param reg: 
 +----------------------------------------------------------------------------*/
void cml_eth_phy_up(unsigned int lan_en, unsigned int wan_en)
{
	unsigned short port;
	
	if(lan_en)
	{
		diag_printf("cml_eth_phy_up...LAN_PORT(%d-%d)\n", PHY_LAN_PORT1, PHY_LAN_PORT4);
		for(port = PHY_LAN_PORT1; port <= PHY_LAN_PORT4; port++)
		{
			cml_eth_phy_pd(port, 0);
		}
	}
	if(wan_en)
	{
		diag_printf("cml_eth_phy_up...WAN_PORT(%d)\n", PHY_WAN_PORT);
		cml_eth_phy_pd(PHY_WAN_PORT, 0);
	}
}
/*!-----------------------------------------------------------------------------
 *  cml_eth_phy_pd:
 *      \brief power down/up of phy.
 *	\param phy: 
 *	\param reg: 
 +----------------------------------------------------------------------------*/
void cml_eth_phy_pd(unsigned short phy, unsigned int pd)
{
	unsigned short val;

	val = cml_eth_mdio_rd(phy, 0); 
	if(pd)
		cml_eth_mdio_wr(phy, 0, val | 0x0800);
	else
		cml_eth_mdio_wr(phy, 0, val & ~0x0800);
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_phy_down:
 *      \brief power down of lan/wan.
 *	\param phy: 
 *	\param reg: 
 +----------------------------------------------------------------------------*/
void cml_eth_phy_down(unsigned int lan_en, unsigned int wan_en)
{
	unsigned short port;

	if(lan_en)
	{
		diag_printf("cml_eth_phy_down...LAN_PORT(%d-%d)\n", PHY_LAN_PORT1, PHY_LAN_PORT4);
		for(port = PHY_LAN_PORT1; port <= PHY_LAN_PORT4; port++)
		{
			cml_eth_phy_pd(port, 1);
		}
	}
	if(wan_en)
	{
		diag_printf("cml_eth_phy_down...WAN_PORT(%d)\n", PHY_WAN_PORT);
		cml_eth_phy_pd(PHY_WAN_PORT, 1);
	}
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_preset_175d
 * \brief disable cpu port for safe
 * \param void
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_preset_175d(void)
{
  	MREG(MCR) = (MCR_INIT&0xff000000);
	cml_eth_mdio_wr(20, 6 , 0x0f3f);
}

struct phycfg
{
	char phyr;
	char reg; 
	unsigned short val;
};

//#define	CONFIG_IP175D_P01
#ifdef	CONFIG_IP175D_P01
static struct phycfg ip175d_cfg[]=
{
	/* indepent vlan learning */
	{20,13,0x9},  /* splite address table into 1K table for unicast frame and 1K table for multicast frame */
	/* tag-based vlan */
	{22,0,0x7ff},
	/* vlan enable */
	{22,10,0x3},	/* valid entry 0-1 */
	/* port0,PVID=1 */
	{22,14,0x1001},
	{22,4,1},
	/* port1,PVID=2 */
	{22,15,0x2002},
	{22,5,2},
	/* vlan1 member: 100010, 00- vlan0 member: 100001 */
	{23,0,0x2221},
	/* port5, add tag vid1,vid0 */
	{23,8,0x2020},
	/* port0~4, remove tag for vid0,vid1 */
	{23,0x10,0x1f1f},
	/* enable wan port forwarding*/
	{20, 6 , 0x3f3f},
	{-1, -1, -1},
};
#else
#if 0
static struct phycfg ip175d_cfg[]=
{
	/* indepent vlan learning */
	{20,13,0x9},  //yf: need check
	/* tag-based vlan */
	{22,0,0x7ff},
	/* vlan enable */
	{22,10,0x3},
	/* port0,PVID=1 */
	{22,14,0x1001},
	{22,4,1},
	{22,5,1},
	{22,6,1},
	{22,7,1},
	/* port1,PVID=1 */
	{22,15,0x2002},
	{22,8,2},
	/* vlan1 member: 110000, vlan0 member: 101111 */
	{23,0,0x032f},
	/* port5, add tag vid1,vid0 */
	{23,8,0x2020},
	/* port0~4, remove tag for vid0,vid1 */
	{23,0x10,0x1f1f},
	/* enable wan port forwarding*/
	{20, 6 , 0x3f3f},
	{-1, -1, -1},
};
#else
static struct phycfg ip175d_cfg[]=
{
	/* indepent vlan learning */
	{20,13,0x9},  /* splite address table into 1K table for unicast frame and 1K table for multicast frame */
	/* tag-based vlan */
	{22,0,0x7ff},
	/* vlan enable */
	{22,10,0x1f},
#ifdef CML_WAN_P0	
	/* port0,PVID=1 */
	{22,14,0x2002},
	{22,4,2},
	/* port1,PVID=1 */
	{22,15,0x1001},
	{22,5,1},
	/* port2,PVID=1 */
	{22,16,0x1001},
	{22,6,1},
	/* port3,PVID=1 */
	{22,17,0x1001},
	{22,7,1},
	/* port4,PVID=2 */
	{22,18,0x1001},
	{22,8,1},
	/* vlan1 member: 111110, vlan0 member: 100001 */
	{23,0,0x3e21},
	/* vlan3 member: 111110, vlan2 member: 111110 */
	{23,1,0x3e3e},
	/* vlan4 member: 111110 */
	{23,2,0x003e},
#ifdef QOS_OUTBOUND_SCHE_WFQ
	/* assign packet from cpu/wan port to output queue 3, others to output queue1 */
	{25, 1, 0x0d57},
#else
	/* assign packet from cpu/wan port to output queue 3, others to output queue0 */
	{25, 1, 0x0c03},
#endif
#else
	/* port0,PVID=1 */
	{22,14,0x1001},
	{22,4,1},
	/* port1,PVID=1 */
	{22,15,0x1001},
	{22,5,1},
	/* port2,PVID=1 */
	{22,16,0x1001},
	{22,6,1},
	/* port3,PVID=1 */
	{22,17,0x1001},
	{22,7,1},
	/* port4,PVID=2 */
	{22,18,0x2002},
	{22,8,2},
	/* vlan1 member: 101111, vlan0 member: 101111 */
	{23,0,0x0b2f},
	/* vlan3 member: 101111, vlan2 member: 101111 */
	{23,1,0x0b2f},
	/* vlan4 member: 110000 */
	{23,2,0x0030},
#ifdef QOS_OUTBOUND_SCHE_WFQ
	/* assign packet from cpu/wan port to output queue 3, others to output queue1 */
	{25, 1, 0x0f55},
#else
	/* assign packet from cpu/wan port to output queue 3, others to output queue0 */
	{25, 1, 0x0f00},
#endif
#endif
#ifdef QOS_OUTBOUND_SCHE_WFQ
	{21, 8, 0x0001},
	{21, 9, 0x00fa},
	{21, 10, 0x007d},
	{21, 12, 0x031d},
#endif
	/* assign port0~port5 schedule type = 2'b11 */
	{25,22,0x0aaa},
	/* assign port base priority enable */
	{25,0,0x003f},
	/* port5, add tag vid1,vid0 */
	{23,8,0x2020},
	/* port5, add tag vid3,vid2 */
	{23,9,0x2020},
	/* port5, add tag vid4 */
	{23,0x0a,0x0020},
	/* port0~4, remove tag for vid0,vid1 */
	{23,0x10,0x1f1f},
	/* port0~4, remove tag for vid2,vid3 */
	{23,0x11,0x1f1f},
	/* port0~4, remove tag for vid4,vid5 */
	{23,0x12,0x1f1f},
	/* enable wan port forwarding*/
	{20, 6 , 0x3f3f},
	{-1, -1, -1},
};
#endif
#endif
static struct phycfg ip175c_cfg[]=
{
	{29,23,0x07c2},
	{29,24,1},	//port0, PVID=1
	{29,25,1},	//port1, PVID=1
	{29,26,1},	//port2, PVID=1
	{29,27,1},	//port3, PVID=1
	{29,28,2},	//port4, PVID=2
	{29,30,2},	//mii0,  PVID=2
	{30, 1,0x2f30},	//VLAN1=mii,port0~port3 VLAN0=mii0,port4
	{30, 2,0x3f30}, //VLAN3=mii,port0~port4 VLAN2=mii0,port4
	{30, 9,0x1089},	//WAN:port4, VLAN:enable, VID:0, Router:enable, LAN: VLAN1
	{29,18,0xfc08},	//flow-control
	{-1, -1, -1},
};
void cml_eth_bridge_enable(void)
{
	/* indepent vlan learning */
	/* tag-based vlan */
	cml_eth_mdio_wr(22,0,0x7ff);
	/* vlan enable */
	cml_eth_mdio_wr(22,10,0x1f);
	/* port0,PVID=1 */
	cml_eth_mdio_wr(22,14,0x1001);
	cml_eth_mdio_wr(22,4,1);
	/* port1,PVID=1 */
	cml_eth_mdio_wr(22,15,0x1001);
	cml_eth_mdio_wr(22,5,1);
	/* port2,PVID=1 */
	cml_eth_mdio_wr(22,16,0x1001);
	cml_eth_mdio_wr(22,6,1);
	/* port3,PVID=1 */
	cml_eth_mdio_wr(22,17,0x1001);
	cml_eth_mdio_wr(22,7,1);
	/* port4,PVID=1*/
	cml_eth_mdio_wr(22,18,0x1001);
	cml_eth_mdio_wr(22,8,1);
	/* vlan1 member: 101111, vlan0 member: 101111 */
	cml_eth_mdio_wr(23,0,0x3f3f);
	/* vlan3 member: 101111, vlan2 member: 101111 */
	cml_eth_mdio_wr(23,1,0x3f3f);
	/* vlan4 member: 110000 */
	cml_eth_mdio_wr(23,2,0x003f);
	/* assign port0~port5 schedule type = 2'b11 */
	cml_eth_mdio_wr(25,22,0x0aaa);
	/* assign port base priority enable */
	cml_eth_mdio_wr(25,0,0x003f);
	/* port5, add tag vid1,vid0 */
	cml_eth_mdio_wr(23,8,0x2020);
	/* port5, add tag vid3,vid2 */
	cml_eth_mdio_wr(23,9,0x2020);
	/* port5, add tag vid4 */
	cml_eth_mdio_wr(23,0x0a,0x0020);
	/* port0~4, remove tag for vid0,vid1 */
	cml_eth_mdio_wr(23,0x10,0x1f1f);
	/* port0~4, remove tag for vid2,vid3 */
	(23,0x11,0x1f1f);
	/* port0~4, remove tag for vid4,vid5 */
	(23,0x12,0x1f1f);
	(20, 6 , 0x3f3f);
	
}	
/*!-----------------------------------------------------------------------------
 *  cml_eth_switch_init:
 * \brief Initialize the switch and configure the VLAN.
 *
 * \param void
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_switch_init(void)
{
	struct phycfg *regset;
	short val;
	// 175D
	if (0x175d == cml_eth_mdio_rd(20,0))
	{
		DBG_MSG(DEBUG_INIT, "Init ip175d switch\n");
		regset = & ip175d_cfg[0];
	}
	else
	{
		DBG_MSG(DEBUG_INIT, "Init ip175c switch\n");
		regset = & ip175c_cfg[0];
	}
	while (regset->phyr!=-1)
	{
		cml_eth_mdio_wr(regset->phyr, regset->reg, regset->val);
		regset++;
	}
	/* Broadcast Storm Protection enable */
	{
		val = cml_eth_mdio_rd(20, 16);
		cml_eth_mdio_wr(20, 16 , val|0x1f00);
		cml_eth_mdio_wr(20, 17 , 0xffff);
		cml_eth_mdio_wr(20, 18 , 0xffff);

		val = cml_eth_mdio_rd(20, 19);
		cml_eth_mdio_wr(20, 19, val|0x0ff);
	}

	/* splite address table into 1K table for unicast frame and 1K table for multicast frame */
	val = cml_eth_mdio_rd(20, 13);
	cml_eth_mdio_wr(20, 13, val|0x8);
	
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_setmac:
 * \brief Set the MAC address for a specified Ethernet interface.
 *
 * \param unit: the number for specified interface
 * \param mac: Pointer to a 6-byte array which contain the address
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_setmac(int unit, unsigned char *mac)
{
	unsigned short idx;
	char fid;
	unsigned int hi = (mac[0] << 8) | mac[1] ;
  	unsigned int lo = (mac[2] << 24) | (mac[3] << 16) | (mac[4] << 8) | mac[5];

	if (unit==0)
	{
		MREG(MSA00) = hi;
  		MREG(MSA01) = lo;
		fid = 1;
	}
	else
	{
		MREG(MSA10) = hi;
  		MREG(MSA11) = lo;
		fid = 2;
	}
	DBG_MSG(DEBUG_INFO, "cml_eth%d: mac %02x:%02x:%02x:%02x:%02x:%02x\n",
			unit, mac[0], mac[1],mac[2], mac[3], mac[4], mac[5]);

	/* configure myself's mac address in switch to avoid mac address land attack */
	cml_eth_macatable_wr((char *)mac, fid, unit, (0x6<<3)|(fid<<6), 0x2);

}

/*!-----------------------------------------------------------------------------
 *  cml_eth_macatable_rd:
 * \brief Get the information from switch address table.
 *
 * \param mac: Pointer to a 6-byte array which contain the address
 * \param fid: the number of fid
 * \param num: entry number
 * \param databuf: data pointer for storing entry information
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_macatable_rd(unsigned char *mac, char fid, char num, unsigned short *databuf)
{
	unsigned short idx;

	if(cml_eth_mdio_rd(20, 13) & 0x8)
	{
		idx = fid ^ mac[0] ^ mac[1] ^ mac[2] ^ mac[3] ^ mac[4] ^ mac[5];
		if(mac[0] & 1)
			idx = idx<<2|num|(1<<10);
		else
			idx = idx<<2|num;
	}
	else
	{
		idx =	(fid<<3 | (mac[0]>>5)) ^ 
		((((short)mac[0]&0x1f) << 4)|((((short)mac[1]) >> 4))) ^  
		((((short)mac[1]&0xf) << 5)|((((short)mac[2]) >> 3))) ^
		((((short)mac[2]&0x7) << 6)|((((short)mac[3]) >> 2))) ^
		((((short)mac[3]&0x3) << 7)|((((short)mac[4]) >> 1))) ^
		((((short)mac[4]&0x1) << 8)|(((short)mac[5])));

		idx = idx<<2|num;
	}
	cml_eth_mdio_wr(21, 14, idx|(2<<11)|(1<<15));
	while((cml_eth_mdio_rd(21, 14) & 0x2000) == 0);

	databuf[0] = cml_eth_mdio_rd(21, 15);
	databuf[1] = cml_eth_mdio_rd(21, 16);
	databuf[2] = cml_eth_mdio_rd(21, 17);
	databuf[3] = cml_eth_mdio_rd(21, 18);
	databuf[4] = cml_eth_mdio_rd(21, 19);
}

void cml_eth_macatable_dump()
{
	unsigned short idx;
	unsigned short databuf[5];

	for(idx=0;idx<2048;idx++)
	{
	cml_eth_mdio_wr(21, 14, idx|(2<<11)|(1<<15));
	while((cml_eth_mdio_rd(21, 14) & 0x2000) == 0);

	databuf[0] = cml_eth_mdio_rd(21, 15);
	databuf[1] = cml_eth_mdio_rd(21, 16);
	databuf[2] = cml_eth_mdio_rd(21, 17);
	databuf[3] = cml_eth_mdio_rd(21, 18);
	databuf[4] = cml_eth_mdio_rd(21, 19);
	diag_printf("idx=%x,mac=%04x%04x%04x, info=%x, flag=%x\n", idx, databuf[0], databuf[1], databuf[2], databuf[3], databuf[4]);
	}
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_macatable_wr:
 * \brief Set the information to switch address table.
 *
 * \param mac: Pointer to a 6-byte array which contain the address
 * \param fid: the number of fid
 * \param num: entry number
 * \param info: reg 21.18
 * \param flag: reg 21.19
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_macatable_wr(unsigned char *mac, char fid, char num, unsigned short info, unsigned short flag)
{
	unsigned short idx;

	if(cml_eth_mdio_rd(20, 13) & 0x8)
	{
		idx = fid ^ mac[0] ^ mac[1] ^ mac[2] ^ mac[3] ^ mac[4] ^ mac[5];
		if(mac[0] & 1)
			idx = idx<<2|num|(1<<10);
		else
			idx = idx<<2|num;
	}
	else
	{
		idx =	(fid<<3 | (mac[0]>>5)) ^ 
		((((short)mac[0]&0x1f) << 4)|((((short)mac[1]) >> 4))) ^  
		((((short)mac[1]&0xf) << 5)|((((short)mac[2]) >> 3))) ^
		((((short)mac[2]&0x7) << 6)|((((short)mac[3]) >> 2))) ^
		((((short)mac[3]&0x3) << 7)|((((short)mac[4]) >> 1))) ^
		((((short)mac[4]&0x1) << 8)|(((short)mac[5])));

		idx = idx<<2|num;
	}
	cml_eth_mdio_wr(21, 15, *(unsigned short *)&mac[4]);
	cml_eth_mdio_wr(21, 16, *(unsigned short *)&mac[2]);
	cml_eth_mdio_wr(21, 17, *(unsigned short *)&mac[0]);
	cml_eth_mdio_wr(21, 18, info);
	cml_eth_mdio_wr(21, 19, flag);

	cml_eth_mdio_wr(21, 14, idx|(1<<11)|(1<<15));
}

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 * cml_eth_hwreset:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
static void cml_eth_hwreset(void)
{
	DBG_MSG(DEBUG_INIT, "cml_eth_hwreset\n");
	MREG(MSRST) = 0xfff;
	MREG(MSRST) = 0;
}
/// @endcond

inline void cml_eth_int_enable(void)
{
	MREG(MIM) &= ~MIM_INIT;
}

inline void cml_eth_int_disable(void)
{
	MREG(MIM) |= MIM_INIT;
}
/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 * cml_eth_tx_start:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
inline void cml_eth_tx_start(void)
{
  MREG(MCR) |= (MCR_TXDMA_EN | MCR_TXMAC_EN);
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 * cml_eth_hwd_init:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
#define	HWD_RDY	(1<<31)
#define	HWD_WR	(1<<28)
#define	HWD_IDX_S 18 	// which entry (0~23)
#define	HWD_HWI_S 16 	//which half-word (0~3)

static void cml_eth_hwd_init(void)
{
	unsigned short i,j;
	unsigned long x;

	for(i=0;i<24;i++)
	{
		for (j=0; j < 4; j++)
		{
			while (HWD_RDY & MREG(HWDESC));
			x = HWD_RDY|HWD_WR |((unsigned long)i<<HWD_IDX_S) | ((unsigned long)j<<HWD_HWI_S) ;
			if (j==0)
			{
				x |= ((unsigned long)DS0_RX_OWNBIT >>16)  |
					((i==23)?  (DS0_END_OF_RING >>16) : 0 ) ;
			}
			//diag_printf("x=%08x\n", x);
			MREG(HWDESC) = x; 
		}
	}
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 * cml_eth_hwinit:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
static void cml_eth_hwinit(void)
{
	DBG_MSG(DEBUG_INIT, "%s\n", __func__);

	cml_eth_hwd_init();

	MREG(MHWDB) = (unsigned int) pcme->hw.prxd;
	MREG(MTXDB) = (unsigned int) pcme->tx.ptxd;
	MREG(MSWDB) = (unsigned int) pcme->rx.prxd;
	
	MREG(MAC_RXCR) = 0x08401004;

#ifdef	CONFIG_FPGA
	MREG(MFP) = 0x14ffff;
#else
	//MREG(MFP) = 0x2dffff;
	MREG(MFP) = 0x10ffff;
#endif

	MREG(MVL01) = 0x10020001;
	MREG(MVL23) = 0x10040003;
	MREG(MVL45) = 0x10060005;
	MREG(MVL67) = 0x10080007;	

	MREG(MCR) = MCR_INIT ;
  	MREG(MRXCR) = MRXCR_INIT ;
  	MREG(MAUTO) = 0x81ff0030 ;
	MREG(MDROPS) = MDROPS_INIT ;
	MREG(ML3) = ML3_INIT ;
	
	MREG(MIC)= MREG(MIS);
	cml_eth_int_enable();
	
   	//using icplus 175d
	DBG_MSG(DEBUG_INIT, "\nInit switch, LAN(VID 1): p0~p3, WAN(VID 2): p4\n");
	cml_eth_switch_init();	
	
	/* HW NAT initialize */
	cml_hnat_init();

}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *  cml_eth_buf_free:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
int cml_eth_buf_free (void)
{
	unsigned long cb,nb,*p;

	if (0 ==(cb = MREG(MBUFB)))
	{
		DBG_MSG(DEBUG_INIT, "%s MBUF=0 can't be free\n",__func__);
		return -1;
	}
	for (;cb;)
	{
		if (3 & (unsigned long)cb )
		{
			DBG_MSG(DEBUG_INIT, "%s unaligned BUF=%08lx, can't be free\n",__func__, cb);
			break;
		}
		p = (unsigned long*)nonca_addr((cb<<8)>>8);
		nb = *p; // next
		eth_drv_free_buf( (void*)( PHYTOVA((cb<<8)>>8)-BUF_HW_OFS)) ;
		cb = nb;	
	}
	return 0;
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *  cml_eth_buf_init:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
int cml_eth_buf_init (void)
{
	unsigned long first, cb, nb;
	int i;

	if (!(cb = first = (unsigned long)eth_drv_alloc_buf()))
	{
		goto no_buf;
	}
	MREG(MBUFB) = virtophy(first)+BUF_HW_OFS ;
	for (i = 0; i < NUM_MBUF-1; i++)
	{
		if (!(nb = (unsigned long)eth_drv_alloc_buf()))
		{
			goto free;
		}
		/* current buf's head point to next buf's head */
		*(unsigned long *)(nonca_addr(cb)+BUF_HW_OFS) = virtophy(nb)+BUF_HW_OFS;
		//saved the pointer (MSB 16bits) of buffer
		//*(unsigned char*)(nonca_addr(cb)+BUF_MSB_OFS) = ((unsigned int)cb)>>24 ;
		cb = nb ;
	}
	*(unsigned long *)(nonca_addr(cb)+BUF_HW_OFS) = 0;
	MREG(MLBUF) = virtophy(cb)+BUF_HW_OFS;
	DBG_MSG(DEBUG_INIT, "%s: MBUF=%x, MLBUF=%x\n", __func__, MREG(MBUFB), MREG(MLBUF));

	return 0;

free:
	cml_eth_buf_free();
no_buf:
	DBG_MSG(DEBUG_INIT, "%s:no buf\n",__func__);
	return -1;
}
/// @endcond


/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *  cml_eth_hw_init:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
int cml_eth_hw_init(void)
{
	pcme = &cml_eth_this;
	
	DBG_MSG(DEBUG_INIT, "%s enter\n", __func__);

	//HAL_DCACHE_FLUSH(g_desc, DESC_SZ);

	cml_eth_preset_175d();
	cml_eth_hwreset();

	pcme->tx.ptxd = (DES *) nonca_addr((unsigned int)g_desc);
	pcme->rx.prxd = (DES *)(pcme->tx.ptxd + NUM_TX_DESC);
	pcme->hw.prxd = (DES *)(pcme->rx.prxd + NUM_RX_DESC);
	
	pcme->tx.tx_bd = &g_txbd[0];
	pcme->tx.tx_max = NUM_TX_DESC;
	pcme->rx.rx_max = NUM_RX_DESC;
	pcme->hw.rx_max = NUM_HW_DESC;
	
	if (cml_eth_buf_init())
		return -1;
	
	cml_eth_txd_init(&pcme->tx);
	cml_eth_rxd_init(&pcme->rx);
	cml_eth_rxd_init(&pcme->hw);

	pcme->irq = CML_ETH_IRQ;
	cml_eth_os_request_irq(CML_ETH_IRQ);

	cml_eth_hwinit();
	
	cml_eth_mdio_wr(4,5,0x480);
	
	return 0;
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *   cml_eth_vif_start:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
void cml_eth_vif_start(cme_vif *vif)
{
	unsigned short which;
	int idx = vif->unit;

	vif = pcme->vifs[idx];

	which = 1 << idx;

	DBG_MSG(DEBUG_INIT, "vif_start%d enter, vifs_map=%x\n", idx, pcme->vifs_map );
	if (pcme->vifs_map & which)
		return;

	if (pcme->vifs_map == 0)
	{
		pcme->sc_main = vif->sc;
		cml_eth_int_enable();
//???	moved out	cyg_drv_interrupt_unmask(pcme->irq);
	}
	pcme->vifs_map |= which;
	DBG_MSG(DEBUG_INIT, "vif_start done: vifs_map=%x\n", pcme->vifs_map);

}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *    cml_eth_rx_stop:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
void cml_eth_rx_stop(RDMA *rdma)
{
	int i;
	DES *prxd;
	void *bufp;
	
  	MREG(MCR) &= ~(MCR_RXDMA_EN | MCR_RXMAC_EN);
  	for (i = 0; i < rdma->rx_max; i++)
  	{
    	prxd = rdma->prxd + i;
    	if(((DS0_RX_OWNBIT | DS0_TX_OWNBIT) >> 16) & *(unsigned short *) prxd)
    	{
    		prxd->w0 &= ~(DS0_RX_OWNBIT | DS0_TX_OWNBIT);
      		prxd->w0 |= DS0_DROP_OF_PACKET;
			if (prxd->w1&0xffffff)
			{
				bufp = (void*)DESC_W1_TO_VBUF(prxd->w1);
				eth_drv_free_buf(bufp);
			}
    	}		
   	}	
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *    cml_eth_tx_stop:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
static void cml_eth_tx_stop(TDMA *tdma)
{
	int i;
	DES *ptxd;
	
	MREG(MCR) &= ~(MCR_TXDMA_EN | MCR_TXMAC_EN);
  	for (i = 0; i < tdma->tx_max; i++)
  	{
    		ptxd = tdma->ptxd + i;
    		if(((DS0_RX_OWNBIT | DS0_TX_OWNBIT) >> 16) & *(unsigned short *) ptxd)
    		{
    			ptxd->w0 &= ~(DS0_RX_OWNBIT | DS0_TX_OWNBIT);
      			ptxd->w0 |= DS0_DROP_OF_PACKET;
    		}
    	}	
}	
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *     cml_eth_vif_stop:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
void cml_eth_vif_stop(cme_vif *vif)
{
	unsigned short which;
	int idx  = vif;

	vif = pcme->vifs[idx];
	which = 1 << idx;

	DBG_MSG(DEBUG_INIT, "vif_stop(%d) vif=%x vifs_map=%x\n", idx, (int) vif, (int) pcme->vifs_map);
	CYG_ASSERTC(vif != 0);

	if (!(pcme->vifs_map & which))
		return;
	
	if (!(pcme->vifs_map &= ~which))
	{
		cml_eth_int_disable();
//??? moved outt		cyg_drv_interrupt_mask(pcme->irq);
#if 0
		cml_eth_rx_stop(&pcme->rx);
		cml_eth_tx_stop(&pcme->tx);
		cml_eth_txd_free(&pcme->tx);
#else
		//DBG_MSG(DEBUG_INIT, "%s Don't free desc currently!\n", __func__);
#endif
		pcme->sc_main = 0;
		return;
	}
	DBG_MSG(DEBUG_INIT, "vif_stop: done vifs_map=%x\n", pcme->vifs_map);

}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *     cml_eth_add_vif:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
int cml_eth_add_vif(int dev_num, cme_vif *vif)
{
	int status = -1;
	
	if (dev_num <0 || dev_num >= MAX_IF_NUM)
		return -1;
	
	if (pcme->vifs[dev_num] == 0)
	{
		pcme->vifs[dev_num] = vif;
		vif->vlan_ctrl = ((dev_num|0x8) << 24);
		status = 0;
	}
	
	return status;
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *     cml_eth_del_vif:
 *
 *      \brief 
 *
 *			\param 
 *
 *      \return 
 +----------------------------------------------------------------------------*/
int cml_eth_del_vif(int dev_num)
{
	cme_vif *vif;

	if (dev_num < 0 || dev_num >= MAX_IF_NUM)
		return -1;

	DBG_MSG(DEBUG_INIT, "cml_eth_del_vif(%d)\n", dev_num);
	
	vif = pcme->vifs[dev_num];
	CYG_ASSERTC(vif != 0);

	cml_eth_setmac(dev_num, vif->macaddr);
	
	pcme->vifs[dev_num] = 0;
	
	return 0;
}
/// @endcond


/*!-----------------------------------------------------------------------------
 *  cml_eth_rx_event:
 * \brief The receive interrupt service handler.
 *
 * \param rdma: pointer of the receive descriptor control.
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_rx_event(RDMA *rdma)
{
	register unsigned int w0, w1;
	unsigned char *nbufp, *bufp;
	unsigned char *pkt;
	DES *prxd;
	short frame_len;
	cme_vif *vif;
	int loop;

	prxd = rdma->prxd;

	DBG_MSG(DEBUG_RXMORE, "rx enter\n");

	prxd = &rdma->prxd[rdma->rx_head];
	w0 = prxd->w0;
	w1 = prxd->w1;
	loop=0;
	while (!(w0 & (DS0_RX_OWNBIT)) && (loop++ < (NUM_RX_DESC/1)))
	{
		DBG_MSG(DEBUG_RXMORE, "rx DS0 w0=%x w1=%x\n",  w0, w1);
		pkt = 0;
		DBG_MSG(DEBUG_RXMORE, "vl=%x\n", RX_VIDX(w1));
		if ((vif = pcme->vifs[RX_VIDX(w1)]) == 0) /* not enable */
		{
			prxd->w1=((w1>>6)<<6) + BUF_HW_OFS;
			goto next;
		}

		if(w0 & DS0_DROP_OF_PACKET)
		{
			if(w0 & (DS0_EIPCS|DS0_EL4CS))
			{
				rdma->rx_err++;
				/*  drop packet  */
				prxd->w1=((w1>>6)<<6) + BUF_HW_OFS;
				goto next;
			}
			else
			{
#if 0				
				diag_printf("Drop!!!, cause:%x\n", RX_ERROR(w0));
				bufp = (unsigned char*)nonca_addr(w1);
				bufp = (unsigned char*)((0xa0<<24) | ((((unsigned int)bufp)<<8)>>8));
				diag_dump_buf_16bit(bufp,64);
#endif				
			}
		}

		frame_len = RX_LEN(w0);
		rdma->rx_byte += frame_len;
		
		if(w0 & DS0_ISBC) 		/* broadcast */
			rdma->rx_bcpkt++;
		else if(w0 & DS0_ISMC)	/* broadcast */
			rdma->rx_mcpkt++;

		rdma->rx_pkt++;
		if ((vif->flags & VIF_ACTIVE) 
			&& (0!=(nbufp=(char*)eth_drv_alloc_buf())) )
		{
			pkt = DESC_W1_TO_VPKT(w1);
			bufp = DESC_W1_TO_VBUF(w1);
			prxd->w1 = (unsigned long)nbufp + BUF_HW_OFS; 
			FLUSH_RX(nbufp, RX_FLUSH_SZ);
		}
		else
		{
			prxd->w1=((w1>>6)<<6) + BUF_HW_OFS;
		}
next:
		if (++rdma->rx_head == rdma->rx_max)
		{
			prxd->w0 = DS0_RX_OWNBIT | DS0_END_OF_RING;
			rdma->rx_head = 0;
		}
		else
			prxd->w0 = DS0_RX_OWNBIT;

		if(pkt != 0)
		{
			void *hw_nat;
    		struct eth_drv_sg sg_list[2];

			//diag_dump_buf_16bit(bufp+0x40,64);
			//printf("hash[0]:%x, hash[1]:%x\n", *(int *)(bufp+4), *(int *)(bufp+12));
				
			hw_nat = cml_hnat_rx_check(bufp, pkt, w0, w1);
			//eth_drv_recv_alt(vif->sc, pkt, frame_len-4, (char**)&bufp, hw_nat); //4byte VLAN
			sg_list[0].buf = pkt;
			sg_list[0].len = frame_len - 4;
			sg_list[1].buf = hw_nat;
#define eth_drv_recv eth_drv_recv_alt	
			if ( eth_drv_recv(vif->sc, &sg_list[0], 1, bufp) ) 
				eth_drv_free_buf(bufp);
		}
		prxd = &rdma->prxd[rdma->rx_head];
		w0 = prxd->w0;
		w1 = prxd->w1;
	}
	
	DBG_MSG(DEBUG_RXMORE, "rx done\n");
}

/*!-----------------------------------------------------------------------------
 *      cml_eth_rxd_free:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/
static void cml_eth_rxd_free(RDMA *rdma)
{
	//TODO 
	// free according to w1's info
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_rxd_init:
 * \brief Initialize the receive buffer descriptors.
 *
 * \param rdma: pointer of the receive descriptor control.
 *
 * \return None
 +----------------------------------------------------------------------------*/
static void cml_eth_rxd_init(RDMA *rdma)
{
	int i;
	DES *prxd;
		
	DBG_MSG(DEBUG_INIT, "cml_eth: init %d rxd @ %08x\n",
			rdma->rx_max, (unsigned int) rdma->prxd);

	prxd = rdma->prxd;
	for (i = rdma->rx_max; i>0; i--, prxd++)
	{
		prxd->w1 = 0;
		prxd->w0 = DS0_RX_OWNBIT;
	}
	(--prxd)->w0 = DS0_END_OF_RING|DS0_RX_OWNBIT;
	rdma->rx_head = rdma->rx_tail = 0;
}

/*!-----------------------------------------------------------------------------
 *  cml_eth_tx_event:
 * \brief The transmit interrupt service handler.
 *
 * \param tdma: pointer of the transmit descriptor control.
 *
 * \return None
 +----------------------------------------------------------------------------*/
void cml_eth_tx_event(TDMA *tdma, int wait_free)
{
	volatile DES *ptxd;
	BDS *tx_bd;
	unsigned int tx_ctrl;
	
	DBG_MSG(DEBUG_TXMORE, "txdone enter\n");

	ptxd = tdma->ptxd;
	tx_ctrl = ptxd[tdma->tx_tail].w0;

#if 1
	if(wait_free && (ptxd[tdma->tx_tail].w0 & DS0_TX_OWNBIT))
	{
		int i = 0;
		while((ptxd[tdma->tx_tail].w0&DS0_TX_OWNBIT))
		{
			diag_printf("%s(%d)\n", __func__, __LINE__);
			i++;
		}
		//diag_printf("i=%d\n", i);
	}
#endif
	while (!(ptxd[tdma->tx_tail].w0 & DS0_TX_OWNBIT) && tdma->tx_free < tdma->tx_max)
	{
		tx_bd = &tdma->tx_bd[tdma->tx_tail];
		DBG_MSG(DEBUG_TX, "TD: %d,%x,%x", tdma->tx_tail,ptxd[tdma->tx_tail].w0, ptxd[tdma->tx_tail].w1);
		DBG_MSG(DEBUG_TX, ",%x\n", (unsigned int)tx_bd->buf);
		if (tx_bd->buf)
		{
			void *buf = tx_bd->buf;
			tdma->tx_ok++;
			tx_bd->txdone?  tx_bd->txdone(buf) : eth_drv_free_buf(buf) ;
			tx_bd->txdone = 0;
			tx_bd->buf = 0;
		}

		if (++tdma->tx_tail == tdma->tx_max)
			tdma->tx_tail = 0;
		tdma->tx_free++;
	}
	DBG_MSG(DEBUG_TXMORE, "txdone finish\n");
}

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *       cml_eth_txd_free:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/
static void cml_eth_txd_free(TDMA *tdma)
{
	short i;
	void *buf;
	BDS *tx_bd = &tdma->tx_bd[0];
	
	for(i=0; i < tdma->tx_max; i++,tx_bd++)
	{
		if (buf=tx_bd->buf)
			(tx_bd->txdone)?  tx_bd->txdone(buf) : eth_drv_free_buf(buf) ;
	}	
}
/// @endcond
/*!-----------------------------------------------------------------------------
 *  cml_eth_txd_init:
 * \brief Initialize the transmit buffer descriptors.
 *
 * \param tdma: pointer of the transmit descriptor control
 *
 * \return None
 +----------------------------------------------------------------------------*/
static void cml_eth_txd_init(TDMA *tdma)
{

	DBG_MSG(DEBUG_INIT, "cml_eth: init %d txd @%08x\n", tdma->tx_max,  (unsigned int) tdma->ptxd);
	/* zero all tx desc */
	bzero((void *) tdma->ptxd, tdma->tx_max * sizeof(DES));
	/* Init buffer descriptor  */
	bzero((void *) tdma->tx_bd, tdma->tx_max * sizeof(BDS));
	
	/*  Marking the last tx descriptor as the end of ring  */
	tdma->ptxd[tdma->tx_max-1].w0 = DS0_END_OF_RING;
	
	tdma->tx_free = tdma->tx_max;
	tdma->tx_head = tdma->tx_tail = 0;
}

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *      cml_eth_rx_start:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/
inline void cml_eth_rx_start(void)
{
  MREG(MCR) |= (MCR_RXDMA_EN | MCR_RXMAC_EN);
}
/// @endcond

/// @cond DOXYGEN
/*!-----------------------------------------------------------------------------
 *      cml_eth_swint:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/
inline void cml_eth_swint(cml_eth *pdev)
{
}
/// @endcond

/*!-----------------------------------------------------------------------------
 *      cml_eth_swint:
 *
 *      \brief 
 *		\param 
 *      \return 
 +----------------------------------------------------------------------------*/

void cml_eth_flowctrl(int onoff)
{
	if(pcme->flowctrl_mode == 0)
		return;
	if(onoff)
	{
		/* send pause frame */
		MREG(MFC1) = 0x2;
		pcme->flowctrl_state = 1;
		diag_printf("send pause!!%x\n", MREG(MFC1));	
	}
	else if(pcme->flowctrl_state)
	{
		/* send clear pause frame */
		MREG(MFC1) = 0x4;
		pcme->flowctrl_state = 0;
		diag_printf("send clear pause!!\n");	
	}
}

/// @endcond
